/* $XConsortium: amigaInit.c,v 5.24 91/08/23 16:14:53 keith Exp $ */
/*-
 * amigaInit.c --
 *	Initialization functions for screen/keyboard/mouse, etc.
 *
 * Copyright (c) 1987 by the Regents of the University of California
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 *
 */

/************************************************************
Copyright 1994 by Eduardo Horvath
Copyright 1987 by Sun Microsystems, Inc. Mountain View, CA.

                    All Rights Reserved

Permission  to  use,  copy,  modify,  and  distribute   this
software  and  its documentation for any purpose and without
fee is hereby granted, provided that the above copyright no-
tice  appear  in all copies and that both that copyright no-
tice and this permission notice appear in  supporting  docu-
mentation,  and  that the names of Sun or MIT not be used in
advertising or publicity pertaining to distribution  of  the
software  without specific prior written permission. Sun and
M.I.T. make no representations about the suitability of this
software for any purpose. It is provided "as is" without any
express or implied warranty.

SUN DISCLAIMS ALL WARRANTIES WITH REGARD TO  THIS  SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
NESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SUN BE  LI-
ABLE  FOR  ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,  DATA  OR
PROFITS,  WHETHER  IN  AN  ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
THE USE OR PERFORMANCE OF THIS SOFTWARE.

********************************************************/

#ifndef	lint
static char sccsid[] = "%W %G Copyright 1987 Sun Micro";
#endif

#include    "amiga.h"
#include    <servermd.h>
#include    "dixstruct.h"
#include    "dix.h"
#include    "opaque.h"
#include    "mipointer.h"

extern int amigaMouseProc();
extern int amigaKbdProc();
extern Bool amigaBW2Probe(), amigaBW2Create();
#ifndef MONO_ONLY
extern Bool amigaCG2CProbe(), amigaCG2CCreate();
extern Bool amigaCG3CProbe(), amigaCG3CCreate();
extern Bool amigaCG4CProbe(), amigaCG4CCreate();
#ifdef FBTYPE_SUNFAST_COLOR /* doesn't exist in amigaos3.x */
extern Bool amigaCG6CProbe(), amigaCG6CCreate();
#endif
#endif
extern void ProcessInputEvents();

extern void SetInputCheck();
extern char *strncpy();
extern GCPtr CreateScratchGC();

#define	XDEVICE	"XDEVICE"
#define	PARENT	"WINDOW_GFX"

static int autoRepeatHandlersInstalled;	/* FALSE each time InitOutput called */

static Bool amigaDevsProbed = FALSE;
Bool amigaSupportsDepth8 = FALSE;
unsigned long amigaGeneration = 0;
int amigaScreenIndex;
Bool FlipPixels = FALSE;


/*-
 *-----------------------------------------------------------------------
 * SigIOHandler --
 *	Signal handler for SIGIO - input is available.
 *
 * Results:
 *	amigaSigIO is set - ProcessInputEvents() will be called soon.
 *
 * Side Effects:
 *	None
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
static void
SigIOHandler(sig, code, scp)
    int		code;
    int		sig;
    struct sigcontext *scp;
{
    amigaEnqueueEvents ();
}

amigaFbDataRec amigaFbData[] = {
#ifndef MONO_ONLY
    amigaBW2Probe,  	"/dev/bwtwo0",	    amigaBW2Create,
    amigaCG3CProbe,  	"/dev/cgthree0",    amigaCG3CCreate,
#ifdef FBTYPE_SUNFAST_COLOR
    amigaCG6CProbe,	"/dev/cgsix0",	    amigaCG6CCreate,
#endif
    amigaCG2CProbe,  	"/dev/cgtwo0",	    amigaCG2CCreate,
    amigaCG4CProbe,  	"/dev/cgfour0",	    amigaCG4CCreate,
#endif
    amigaBW2Probe,  	"/dev/bwtwo0",	    amigaBW2Create,
};

#ifdef MONO_ONLY
#define DEV_START   0
#define DEV_END	    0
#else
#define DEV_START   1
#define DEV_END	    1
#endif

/*
 * NUMSCREENS is the number of supported frame buffers (i.e. the number of
 * structures in amigaFbData which have an actual probeProc).
 */
#define NUMSCREENS (sizeof(amigaFbData)/sizeof(amigaFbData[0]))
#define NUMDEVICES  2

fbFd amigaFbs[NUMDEVICES];

static PixmapFormatRec	formats[] = {
    1, 1, BITMAP_SCANLINE_PAD,	/* 1-bit deep */
    8, 8, BITMAP_SCANLINE_PAD,	/* 8-bit deep */
};
#define NUMFORMATS	(sizeof formats)/(sizeof formats[0])

/*-
 *-----------------------------------------------------------------------
 * amigaNonBlockConsoleOff --
 *	Turn non-blocking mode on the console off, so you don't get logged
 *	out when the server exits.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
amigaNonBlockConsoleOff(arg)
    char	*arg;
{
    register int i;

    i = fcntl(2, F_GETFL, 0);
    if (i >= 0)
	(void) fcntl(2, F_SETFL, i & ~FNDELAY);
}

/*-
 *-----------------------------------------------------------------------
 * InitOutput --
 *	Initialize screenInfo for all actually accessible framebuffers.
 *	The
 *
 * Results:
 *	screenInfo init proc field set
 *
 * Side Effects:
 *	None
 *
 *-----------------------------------------------------------------------
 */

InitOutput(pScreenInfo, argc, argv)
    ScreenInfo 	  *pScreenInfo;
    int     	  argc;
    char    	  **argv;
{
    int     	  i, n, dev;
    int		  nonBlockConsole = 0;
    static int	  setup_on_exit = 0;
    int		  devStart = DEV_START;
    extern Bool	  RunFromSmartParent;

    if (!monitorResolution)
	monitorResolution = 90;
    if (RunFromSmartParent)
	nonBlockConsole = 1;
    for (i = 1; i < argc; i++) {
	if (!strcmp(argv[i],"-debug"))
	    nonBlockConsole = 0;
	else if (!strcmp(argv[i],"-mono"))
	    devStart = 0;
    }
    /*
     *	Writes to /dev/console can block - causing an
     *	excess of error messages to hang the server in
     *	deadlock.  So.......
     */
    if (nonBlockConsole) {
	if (!setup_on_exit) {
	    if (atexit(amigaNonBlockConsoleOff))
		ErrorF("InitOutput: can't register NBIO exit handler\n");
	    setup_on_exit = 1;
	}
	i = fcntl(2, F_GETFL, 0);
	if (i >= 0)
	    i = fcntl(2, F_SETFL, i | FNDELAY);
	if (i < 0) {
	    perror("fcntl");
	    ErrorF("InitOutput: can't put stderr in non-block mode\n");
	}
    }
    pScreenInfo->imageByteOrder = IMAGE_BYTE_ORDER;
    pScreenInfo->bitmapScanlineUnit = BITMAP_SCANLINE_UNIT;
    pScreenInfo->bitmapScanlinePad = BITMAP_SCANLINE_PAD;
    pScreenInfo->bitmapBitOrder = BITMAP_BIT_ORDER;

    pScreenInfo->numPixmapFormats = NUMFORMATS;
    for (i=0; i< NUMFORMATS; i++)
        pScreenInfo->formats[i] = formats[i];

    autoRepeatHandlersInstalled = FALSE;

    if (!amigaDevsProbed)
    {
	n = 0;
	for (i = NUMSCREENS, dev = devStart; --i >= DEV_END; dev++) {
	    if ((*amigaFbData[dev].probeProc)(pScreenInfo, n, dev, argc, argv))
		n++;
	    else
		amigaFbData[dev].createProc = NULL;
	}
	amigaDevsProbed = TRUE;
	if (n == 0)
	    return;
    }
    if (!amigaSupportsDepth8)
	pScreenInfo->numPixmapFormats--;
    for (i = NUMSCREENS, dev = devStart; --i >= DEV_END; dev++) {
	if (amigaFbData[dev].createProc)
	    (*amigaFbData[dev].createProc)(pScreenInfo, argc, argv);
    }
    amigaInitCursor();
    signal(SIGWINCH, SIG_IGN);
}

/*-
 *-----------------------------------------------------------------------
 * InitInput --
 *	Initialize all supported input devices...what else is there
 *	besides pointer and keyboard?
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Two DeviceRec's are allocated and registered as the system pointer
 *	and keyboard devices.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
InitInput(argc, argv)
    int     	  argc;
    char    	  **argv;
{
    DevicePtr p, k;
    static int  zero = 0;
    
    p = AddInputDevice(amigaMouseProc, TRUE);
    k = AddInputDevice(amigaKbdProc, TRUE);
    if (!p || !k)
	FatalError("failed to create input devices in InitInput");

    SetTimeSinceLastInputEvent();
    RegisterPointerDevice(p);
    RegisterKeyboardDevice(k);
    miRegisterPointerDevice(screenInfo.screens[0], p);
    if (!mieqInit (k, p))
	return FALSE;
    signal(SIGIO, SigIOHandler);
}


static Bool
amigaCloseScreen (i, pScreen)
    int		i;
    ScreenPtr	pScreen;
{
    SetupScreen(pScreen);
    Bool    ret;

    signal (SIGIO, SIG_IGN);
    amigaDisableCursor (pScreen);
    pScreen->CloseScreen = pPrivate->CloseScreen;
    ret = (*pScreen->CloseScreen) (i, pScreen);
    (void) (*pScreen->SaveScreen) (pScreen, SCREEN_SAVER_OFF);
    xfree ((pointer) pPrivate);
    return ret;
}

Bool
amigaSaveScreen (pScreen, on)
    ScreenPtr	pScreen;
    int		on;
{
    int		state;

    if (on == SCREEN_SAVER_FORCER)
	SetTimeSinceLastInputEvent();
    else
    {
/*
	if (on == SCREEN_SAVER_ON)
	  (void) ioctl(amigaFbs[pScreen->myNum].fd, VIEW_DISPLAY, 0);
	else
	  (void) ioctl(amigaFbs[pScreen->myNum].fd, VIEW_REMOVE, 0);
*/
    }
    return( TRUE );
}

/*-
 *-----------------------------------------------------------------------
 * amigaScreenInit --
 *	Things which must be done for all types of frame buffers...
 *	Should be called last of all.
 *
 * Results:
 *	TRUE if successful, else FALSE
 *
 * Side Effects:
 *	Both a BlockHandler and a WakeupHandler are installed for the
 *	first screen.  Together, these handlers implement autorepeat
 *	keystrokes on the Sun.
 *
 *-----------------------------------------------------------------------
 */

Bool
amigaScreenAllocate (pScreen)
    ScreenPtr	pScreen;
{
    amigaScreenPtr    pPrivate;

    if (amigaGeneration != serverGeneration)
    {
	amigaScreenIndex = AllocateScreenPrivateIndex();
	if (amigaScreenIndex < 0)
	    return FALSE;
	amigaGeneration = serverGeneration;
    }
    pPrivate = (amigaScreenPtr) xalloc (sizeof (amigaScreenRec));
    if (!pPrivate)
	return FALSE;

    pScreen->devPrivates[amigaScreenIndex].ptr = (pointer) pPrivate;
    return TRUE;
}

Bool
amigaScreenInit (pScreen)
    ScreenPtr	pScreen;
{
    SetupScreen(pScreen);
    extern void   amigaBlockHandler();
    extern void   amigaWakeupHandler();
    static ScreenPtr autoRepeatScreen;
    extern miPointerScreenFuncRec   amigaPointerScreenFuncs;

    pPrivate->installedMap = 0;
    pPrivate->CloseScreen = pScreen->CloseScreen;
    pScreen->CloseScreen = amigaCloseScreen;
    pScreen->SaveScreen = amigaSaveScreen;

    /*
     *	Block/Unblock handlers
     */
    if (autoRepeatHandlersInstalled == FALSE) {
	autoRepeatScreen = pScreen;
	autoRepeatHandlersInstalled = TRUE;
    }

    if (pScreen == autoRepeatScreen) {
        pScreen->BlockHandler = amigaBlockHandler;
        pScreen->WakeupHandler = amigaWakeupHandler;
    }

    if (!amigaCursorInitialize (pScreen))
	miDCInitialize (pScreen, &amigaPointerScreenFuncs);

    amigaInitCursor ();

    return TRUE;
}

extern char *getenv();

/*-
 *-----------------------------------------------------------------------
 * nthdev --
 *	Return the nth device in a colon-separated list of devices.
 *	n is 0-origin.
 *
 * Results:
 *	A pointer to a STATIC string which is the device name.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------
 */
static char *
nthdev (dList, n)
    register char    *dList;	    /* Colon-separated device names */
    int	    n;	  	    /* Device number wanted */
{
    char *result;
    static char returnstring[100];

    while (n--) {
	while (*dList && *dList != ':') {
	    dList++;
	}
	if (*dList)
	    ++dList;
    }
    if (*dList) {
	register char *cp = dList;

	while (*cp && *cp != ':') {
	    cp++;
	}
	result = returnstring;
	strncpy (result, dList, cp - dList);
	result[cp - dList] = '\0';
    } else {
	result = (char *)0;
    }
    return (result);
}

#if 0
/*-
 *-----------------------------------------------------------------------
 * amigaOpenFrameBuffer --
 *	Open a frame buffer according to several rules. If running under
 *	overview and we're set up for it, use the device given in the
 *	PARENT envariable and note that the screen is under overview.
 *	Else find the device to use by looking in the amigaFbData table,
 *	an XDEVICE envariable, a -dev switch or using /dev/fb if trying
 *	to open screen 0 and all else has failed.
 *
 * Results:
 *	The fd of the framebuffer.
 *
 * Side Effects:
 *
 *-----------------------------------------------------------------------
 */
int
amigaOpenFrameBuffer(expect, pfbType, index, fbNum, argc, argv)
    int	    	  expect;   	/* The expected type of framebuffer */
    struct fbtype *pfbType; 	/* Place to store the fb info */
    int	    	  fbNum;    	/* Index into the amigaFbData array */
    int	    	  index;    	/* Screen index */
    int	    	  argc;	    	/* Command-line arguments... */
    char	  **argv;   	/* ... */
{
    char       	  *name=(char *)0;
    int           i;	    	/* Index into argument list */
    int           fd = -1;	    	/* Descriptor to device */
    static int	  devFbUsed=FALSE;  /* true if /dev/fb has been used for a */
    	    	  	    	    /* screen already */
    static Bool	  inited = FALSE;
    static char	  *xdevice; 	/* string of devices to use from environ */
    static char	  *devsw;   	/* string of devices from args */
    struct fbgattr  fbattr;
    int		  type;

    amigaFbs[index].parent = FALSE;

    if (!inited) {
	xdevice = devsw = (char *)NULL;

	xdevice = getenv (XDEVICE);
	/*
	 * Look for an argument of the form -dev <device-string>
	 * If such a one is found place the <device-string> in devsw.
	 */
	for (i = 1; i < argc; i++) {
	    if ((strcmp(argv[i], "-dev") == 0) && (i + 1 < argc)) {
		devsw = argv[i+1];
		break;
	    }
	}
	inited = TRUE;
    }

    /*
     * Attempt to find a file name for the frame buffer 
     */

    /*
     * First see if any device was given on the command line.
     * If one was and the device is both readable and writeable,
     * set 'name' to it, else set it to NULL.
     */
    if (devsw == (char *)NULL ||
	(name = nthdev (devsw, index)) == (char *)NULL ||
	(access (name, R_OK | W_OK) != 0) ||
	(strcmp(name, amigaFbData[fbNum].devName) != 0)) {
	    name = (char *)NULL;
    }
	    
    /*
     * If we still don't have a device for this screen, check the
     * environment variable for one. If one was given, stick its
     * path in name and check its accessibility. If it's not
     * properly accessible, then reset the name to NULL to force the
     * checking of the amigaFbData array.
     */
    if (devsw == (char *)NULL && name == (char *)NULL &&
	xdevice != (char *)NULL &&
	(name = nthdev(xdevice, index)) != (char *)NULL &&
	(access (name, R_OK | W_OK) != 0)) {
	    name = (char *)NULL;
    }

    /*
     * Take the device given in the frame buffer description
     * and see if it exists and is accessible. If it does/is,
     * we will use it, as long as no other device was given.
     */
    if (devsw == (char *)NULL && name == (char *)NULL &&
	access(amigaFbData[fbNum].devName, (R_OK | W_OK)) == 0) {
	    name = amigaFbData[fbNum].devName;
    }

    /*
     * If we still have nothing and have yet to use "/dev/fb" for
     * a screen, default the name to be "/dev/fb"
     */
    if (devsw == (char *)NULL && name == (char *)NULL && !devFbUsed) {
	name = "/dev/fb";
    }


    if (name != (char *) NULL && amigaUseSunWindows()) {
	ErrorF("Not configured to run inside SunWindows\n");
	fd = -1;
    } else if (name) {
	fd = open(name, O_RDWR, 0);
        if (fd < 0) {
	    return (-1);
	} 
	if (ioctl(fd, FBIOGATTR, &fbattr) < 0)
	{
	    if (ioctl(fd, FBIOGTYPE, pfbType) < 0) {
	    	perror("amigaOpenFrameBuffer");
	    	(void) close(fd);
	    	return (-1);
	    }
	    type = pfbType->fb_type;
	}
	else
	{
	    int	i;

	    type = fbattr.real_type;
	    *pfbType = fbattr.fbtype;
	    if (expect == FBTYPE_SUN2BW && expect != type)
	    {
	    	for (i = 0; i < FB_ATTR_NEMUTYPES; i++)
	    	{
		    if (expect == fbattr.emu_types[i])
		    {
		    	type = fbattr.emu_types[i];
			break;
		    }
	    	}
	    }
	}
	/* XXX - this is temporary 'cos the CG4 pretends its a BW2 */
	if (strcmp(name, amigaFbData[fbNum].devName) != 0 && type != expect)
	{
	    (void) close(fd);
	    return (-1);
	}
    }

    if (name && strcmp (name, "/dev/fb") == 0) {
	devFbUsed = TRUE;
    }

    return (fd);
}

#endif
